""" Classes and utilities for connecting to and querying the ALMA metadata database.
Much of this code is swiped or adapted from archive.pyat.pyat.schema.__init__ """

import http
import logging
import os
from pathlib import Path

# pylint: disable=E0401, E0402, R0201, R0903, R0914, W1203

import cx_Oracle
import requests
import sqlalchemy
from sqlalchemy.orm import sessionmaker
from bs4 import BeautifulSoup

from pycapo import CapoConfig

from .errors import (
    NGASFetchError,
    FileErrorException,
    SizeMismatchException,
    NGASServiceErrorException,
)
from .locations_report import NgasFile
from .ngas_retriever import NgasRequester, check_result
from .utilities import LOG_FORMAT


ALMA_CAPO_PREFIX = "almaMetadataDatabase.jdbc"
NGAS_NAASC = ALMA_CAPO_PREFIX

logger = logging.getLogger(__name__)
formatter = logging.Formatter(LOG_FORMAT)
handler = logging.StreamHandler()
handler.setFormatter(fmt=formatter)
logger.addHandler(handler)


class AlmaDbConnector:
    """ for connecting to ALMA's Oracle metadata database """

    def __init__(self):
        set_oracle_env()

    def create_engine(self):
        """
        Create the SQLAlchemy engine.

        Until cx_Oracle can be included in Docker containers, this can be
        done only locally.

        :return:
        """
        capo_config = CapoConfig()
        user = capo_config[ALMA_CAPO_PREFIX + "Username"]
        passwd = capo_config[ALMA_CAPO_PREFIX + "Password"]
        url = capo_config[ALMA_CAPO_PREFIX + "Url"]

        hostport, service_name = url.split("/")[-2:]
        host, port = hostport.split(":")
        dsn = cx_Oracle.makedsn(host, port, service_name=service_name)
        archiveurl = f"oracle://{user}:{passwd}@{dsn}"

        return sqlalchemy.create_engine(archiveurl)

    def create_session(self):
        """
        Create the SQLAlchemy Oracle session.

        :return:
        """
        session_mkr = sessionmaker(self.create_engine())
        return session_mkr()


class AlmaNgasStreamer(NgasRequester):
    """ AlmaFetcher will use this to stream ALMA BDFs from NGAS """

    def send_request(self, destination: Path, ngas_file: NgasFile) -> Path:
        """
        Request this ALMA product from NAASC NGAS.

        :param destination: path to output file
        :param ngas_file: product to fetch
        :return:
        """

        download_url = "http://" + ngas_file.server.server + "/RETRIEVE"
        logger.debug(
            f"attempting streaming download:\nurl: {download_url}\ndestination: {destination}"
        )

        params = {
            "file_id": ngas_file.ngas_file_id,
            "file_version": ngas_file.version,
        }

        try:
            with requests.Session() as session:
                response = session.get(download_url, params=params, stream=True)
                with open(destination, "wb") as file_to_write:
                    for chunk in response.iter_content(chunk_size=8192):
                        file_to_write.write(chunk)
                expected_size = ngas_file.size
                actual_size = os.path.getsize(destination)
                if actual_size == 0:
                    raise FileErrorException(f"{Path(destination).name} was not retrieved")
                if actual_size != expected_size:
                    raise SizeMismatchException(
                        f"expected {Path(destination).name} "
                        f"to be {expected_size} bytes; "
                        f"was {actual_size} bytes"
                    )
        except requests.exceptions.ConnectionError as c_err:
            raise NGASServiceErrorException(
                f"problem connecting with {ngas_file.server.server}"
            ) from c_err
        except AttributeError as a_err:
            logger.warning(f"possible problem streaming: {a_err}")

        if response.status_code != http.HTTPStatus.OK:
            logger.error(f"NGAS does not like this request:\n{response.url}")
            soup = BeautifulSoup(response.text, "lxml-xml")
            ngams_status = soup.NgamsStatus.Status
            message = ngams_status.get("Message")

            raise NGASServiceErrorException(
                {
                    "status_code": response.status_code,
                    "url": response.url,
                    "reason": response.reason,
                    "message": message,
                }
            )

        if check_result(ngas_file, destination):
            logger.debug(f"retrieved {destination} from {ngas_file.server.server}")
            return destination

        raise NGASFetchError(str(destination))


def set_oracle_env():
    """
    Set environment variables required by cx_Oracle.

    :return:
    """
    # TODO: once cx_Oracle is available in the Docker container, swap in
    #  correct values.

    home = Path(os.environ["HOME"])
    instant_client_pathname = "Downloads/instantclient_12_1"

    if "ORACLE_HOME" not in os.environ.keys():
        oracle_home = home / instant_client_pathname
        os.environ["ORACLE_HOME"] = str(oracle_home)

    os.environ["PATH"] = str(os.environ["PATH"]) + f":{instant_client_pathname}"
