""" Retrieval of SDM table XML from ALMA metadata database """

# pylint: disable=E0401, E0402, R0903, W0212

import warnings
from pathlib import Path

from sqlalchemy import exc as sa_exc
from .alma_db_utils import AlmaDbConnector, set_oracle_env
from .errors import FileErrorException
from .locations_report import OracleXml


class OracleXmlRetriever:
    """ Grabs the XML from an ALMA XML_%ENTITIES table """

    def __init__(self):
        set_oracle_env()

    def retrieve(self, sdm: OracleXml) -> Path:
        """
        Get the XML for this archive_uid from this table

        :param sdm: XML table retrieved from ALMA DB
        :return:
        """
        with warnings.catch_warnings():
            # suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)

            engine = AlmaDbConnector().create_engine()
            cursor = engine
            try:
                result = cursor.execute(
                    f"SELECT (xml).getClobVal() FROM ALMA.{sdm.table} "
                    f"WHERE "
                    f"archive_uid = :archive_uid",
                    dict(archive_uid=sdm.archive_uid),
                )
                for row in result:
                    xml = row._row[0]
                    return self.write_to_disk(sdm, xml)

                raise FileErrorException(f"{sdm.table} XML not found for{sdm.archive_uid}")

            finally:
                engine.dispose()

    def write_to_disk(self, sdm: OracleXml, xml: str) -> Path:
        """
        Write the XML for this SDM to disk. Assumption: writing to current dir

        :param self:
        :param xml:
        :return:
        """

        # Filename will be like "CalPointingTable.xml" or "SBSummaryTable.xml";
        # we use the relative_path
        filename = sdm.relative_path
        file = Path(filename)
        with open(file, "w") as out:
            out.write(xml)

        return file
