"""
This module concerns the Fetcher facilities. There are expected to be about
three of these:

 - NgasDirectCopyFetcher, which uses the NGAS direct copy plugin,
 - NgasStreamingFileFetcher, which uses the NGAS streaming mode, and
 - AlmaSdmFetcher, which queries the ALMA database to fetch something

Each of these is fairly stupid and just tries to fetch its one file, one time.
Retrying is not handled here, the RetryableFileFetcher is used for that.
"""
import time
from enum import Enum
from pathlib import Path

# pylint: disable=E0401, E0402, W0221

import cx_Oracle
import sqlalchemy
from pycapo import CapoConfig

from .exceptions import (
    RetryableFetchError,
    UnsatisfiableConstraint,
    NormalModeFileExistsError,
)
from .interfaces import FileFetcher, LocatedFile
from .locations import NgasFile, OracleXml


class ForceMode(Enum):
    """
    ForceMode is either normal or force. In normal mode, we refuse to overwrite an
    existing file. In force mode, we are more than willing to destroy your old data.
    """

    NORMAL = False
    FORCE = True

    def check_destination(self, path: Path):
        if self == self.NORMAL and path.exists():
            raise NormalModeFileExistsError(
                f"Cannot fetch to {path} because the file already exists."
            )


class NgasDirectCopyFileFetcher(FileFetcher):
    """
    Fetch a file using the NGAS direct copy mechanism.
    """

    def __init__(self, file: NgasFile, force=ForceMode.NORMAL):
        super().__init__()
        self._file = file
        self.force = force

    @property
    def file(self) -> NgasFile:
        return self._file

    def shares_resources_with(self, other: "FileFetcher") -> bool:
        return self.file.shares_resources_with(other.file)

    def do_fetch(self) -> Path:
        """
        Fetch a file from NGAS via direct copy, or die trying.

        :return: the fetched file
        """
        # ensure that we actually can fetch
        if not self.file.can_direct_copy():
            raise UnsatisfiableConstraint(f"Direct copy to {self.file.destination} is not possible")

        # check the destination
        self.force.check_destination(self.file.destination)

        # create the directory for this file
        self.file.prepare_destination()

        # direct copy the file
        self.file.direct_copy()

        return self.file.destination

    def __str__(self):
        return f"NGAS Direct Copy of {self.file}"


class NgasStreamingFileFetcher(FileFetcher):
    """
    Fetch a file using NGAS streaming.
    """

    def __init__(self, file: NgasFile, force=ForceMode.NORMAL):
        super().__init__()
        self._file = file
        self.force = force

    @property
    def file(self) -> NgasFile:
        return self._file

    def shares_resources_with(self, other: "FileFetcher") -> bool:
        return self.file.shares_resources_with(other.file)

    def do_fetch(self) -> Path:
        """
        Fetch a file from NGAS via streaming, or die trying.

        :return: the fetched file
        """
        # check the destination
        self.force.check_destination(self.file.destination)

        # create the directory for this file
        self.file.prepare_destination()

        # run the streaming fetch
        self.file.stream()

        return self.file.destination

    def __str__(self):
        return f"NGAS streaming copy of {self.file}"


class OracleXmlFetcher(FileFetcher):
    """
    Fetch a file by retrieving it from the ALMA database.
    """

    ALMA_CAPO_PREFIX = "almaMetadataDatabase.jdbc"

    engine = None

    def __init__(self, file: OracleXml, force=ForceMode.NORMAL):
        super().__init__()
        self._file = file
        self.force = force

    @property
    def file(self) -> OracleXml:
        return self._file

    def shares_resources_with(self, other: "FileFetcher") -> bool:
        return self.file.shares_resources_with(other.file)

    def do_fetch(self) -> Path:
        """
        Retrieve an ALMA SDM table as XML from the ALMA DB.

        :return: the file we got
        """
        # check the destination
        self.force.check_destination(self.file.destination)

        # create the directory for this file
        self.file.prepare_destination()

        engine = self.get_engine()
        try:
            result = list(
                engine.execute(
                    f"SELECT (xml).getClobVal() FROM ALMA.{self.file.table} WHERE "
                    f"archive_uid = :archive_uid",
                    dict(archive_uid=self.file.archive_uid),
                )
            )[0][0]

            Path(self.file.destination).write_text(result)
        except Exception as ex:
            raise RetryableFetchError("Something went wrong trying to fetch") from ex
        finally:
            engine.dispose()

        return self.file.destination

    @classmethod
    def get_engine(cls):
        """
        Get an engine against which queries can be run.

        :return:
        """
        if cls.engine is None:
            capo_config = CapoConfig()
            user = capo_config[cls.ALMA_CAPO_PREFIX + "Username"]
            passwd = capo_config[cls.ALMA_CAPO_PREFIX + "Password"]
            url = capo_config[cls.ALMA_CAPO_PREFIX + "Url"]

            hostport, service_name = url.split("/")[-2:]
            host, port = hostport.split(":")
            dsn = cx_Oracle.makedsn(host, port, service_name=service_name)
            archiveurl = f"oracle://{user}:{passwd}@{dsn}"

            cls.engine = sqlalchemy.create_engine(archiveurl)
        return cls.engine

    def __str__(self):
        return f"Oracle XML fetch of {self.file}"


class FileFetcherDecorator(FileFetcher):
    """
    Decorator to allow us to interpose other behavior into a FileFetcher.
    """

    def __init__(self, underlying: FileFetcher):
        super().__init__()
        self.underlying = underlying

    @property
    def file(self) -> LocatedFile:
        return self.underlying.file

    def shares_resources_with(self, other: "FileFetcher") -> bool:
        return self.underlying.shares_resources_with(other)

    def do_fetch(self) -> Path:
        return self.underlying.do_fetch()

    def __str__(self):
        return str(self.underlying)


class RetryableFileFetcher(FileFetcherDecorator):
    """A FileFetcher that retries N times in the event of failure, waiting
    longer each time"""

    def __init__(self, underlying: FileFetcher, retries=3):
        super().__init__(underlying)
        self.retries = retries

    def do_fetch(self, attempt=1) -> Path:
        try:
            return self.underlying.do_fetch()
        except RetryableFetchError as r_err:
            if attempt < self.retries:
                # sleep for 2, 4, 8 seconds between attempts
                time.sleep(2 ** attempt)
                print(f"{self.underlying} (attempt #{attempt+1})")
                return self.do_fetch(attempt + 1)
            raise r_err

    def __str__(self):
        return f"{self.underlying} (with up to {self.retries} retries)"


class DryRunFakeFileFetcher(FileFetcher):
    """
    A fake fetcher we can use to implement dry runs and other tests
    """

    def __init__(self, file: LocatedFile, force=False):
        super().__init__()
        self._file = file
        self.force = force

    @property
    def file(self) -> LocatedFile:
        return self._file

    def do_fetch(self) -> Path:
        """
        In a dry run, we don't actually do any fetching.
        """

    def shares_resources_with(self, other: "FileFetcher") -> bool:
        return self.file.shares_resources_with(other.file)

    def __str__(self):
        return f"Fake fetch of {self.file}{' (forced)' if self.force else ''}"


class NgasModeDetectingFetcher(FileFetcherDecorator):
    """
    This fetcher is used when the user has no preference between streaming and direct-copy.
    The fetcher itself looks at the file and the destination and makes a determination about
    which one to use.
    """

    def __init__(self, file: NgasFile, force=ForceMode.NORMAL):
        # If the user has no preference, they would prefer direct copy
        # if it is possible because it's faster
        if file.can_direct_copy():
            super().__init__(NgasDirectCopyFileFetcher(file, force=force))
        else:
            super().__init__(NgasStreamingFileFetcher(file, force=force))
