"""
Abstracts the work of talking to NGAS. Clients of NGAS should not have to know
a lot about the way NGAS requests work, especially since there are just a couple
things you can do with a request (stream a file to disk or direct copy a file
to disk).

NgasConnection manages that for you; all you need to bring to the table is the
file ID, the file version, and where you want the file written to disk.

In principle it should be less expensive to call these multiple times reusing
the same connection, but in practice I doubt the savings is meaningful, and the
current product-fetcher design does not reuse the connection anyway.
"""

# pylint: disable=E0401, E0402

import http
from contextlib import AbstractContextManager
from pathlib import Path
from xml.etree import ElementTree

import requests

from .exceptions import RetryableFetchError, FetchError, NgasServiceErrorException

SERVICE_EXC_TEXT = "Service exception"


class NgasConnection(AbstractContextManager):
    """
    This class makes it easier to deal with NGAS connections.
    """

    def __init__(self, server):
        self.session = requests.Session()
        self.download_url = f"http://{server}/RETRIEVE"

    def stream(self, file_id: str, file_version: int, destination: Path):
        """
        Stream the request to disk. Throws an exception if it fails.

        :param file_id:  the file to stream
        :param file_version:   the version of the file to request
        :param destination:   where to place the result
        :return: None
        """
        # this is the default behavior of NGAS, so we just make a web request and
        # it should be fine
        params = {
            "file_id": file_id,
            "file_version": file_version,
        }

        response = self.get(params=params, stream=True)
        with open(destination, "wb") as file_to_write:
            for chunk in response.iter_content(chunk_size=8192):
                file_to_write.write(chunk)

    def direct_copy(self, file_id, file_version, destination):
        """
        Direct copy a file to a destination. Throws an exception if it fails.

        :param file_id:  the file to copy
        :param file_version:  the version of the file to request
        :param destination:  where to place the result
        :return:
        """

        # the trick here is to just specify the output file when providing the direct copy
        # plugin
        params = {
            "file_id": file_id,
            "file_version": file_version,
            "processing": "ngamsDirectCopyDppi",
            "processingPars": "outfile=" + str(destination.absolute()),
        }

        self.get(params=params)

    def get(self, **kwargs) -> requests.Response:
        """
        Perform an NGAS request. The keyword arguments flow through to the
        underlying requests.get().

        :param kwargs:  additional arguments to the session.get()
        :return: a response object
        """
        # run the request
        response = self.session.get(self.download_url, **kwargs)

        # check the response and make sure it's 200 OK
        # if it is not, we parse out the NGAMS Status Message, which
        # feeds an exception we throw about how we couldn't do this request
        if response.status_code != http.HTTPStatus.OK:
            doc = ElementTree.fromstring(response.text)
            ngams_status = doc.find("Status").attrib["Message"]

            ex = NgasServiceErrorException(
                {
                    "status_code": response.status_code,
                    "url": response.url,
                    "reason": response.reason,
                    "message": ngams_status,
                }
            )

            # No point in retrying a 404 Not Found
            if response.status_code == http.HTTPStatus.NOT_FOUND:
                raise FetchError(SERVICE_EXC_TEXT, ex)

            # otherwise, maybe, yeah
            raise RetryableFetchError(SERVICE_EXC_TEXT, ex)
        return response

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()
