""" Launches NGAS retrievers in parallel """

# pylint: disable=E0401, E0402
import copy
import logging
from pathlib import Path
from typing import List, Dict

# pylint: disable=E0402, E0611, R0801, W1203
from datafetcher.errors import FileErrorException
from datafetcher.oracle_xml_retriever import OracleXmlRetriever

from .ngas_retriever import NgasRetriever, NgasNonRetriever
from .utilities import LOG_FORMAT

from .locations_report import NgasTarget, OracleXml

logger = logging.getLogger(__name__)
formatter = logging.Formatter(LOG_FORMAT)
handler = logging.StreamHandler()
handler.setFormatter(fmt=formatter)
logger.addHandler(handler)


class ParallelRetrieverLauncher:
    """Launches one or more OracleXmlRetrievers and one or more NgasRetrievers """

    def __init__(
        self,
        ngas_targets: List[NgasTarget],
        alma_sdms: List[OracleXml],
        dry_run: bool,
        force_overwrite: bool,
    ):
        self.ngas_targets = ngas_targets
        self.alma_sdms = alma_sdms
        self.files_by_server = self.organize_file_info()
        self.dry_run = dry_run
        self.force_overwrite = force_overwrite

    def organize_file_info(self) -> Dict:
        """The location report we get back looks like this, for each file:

        {"ngas_file_id":"17B-197_2018_02_19_T15_59_16.097.tar",
        "subdirectory":"17B-197.sb34812522.eb35115211.58168.58572621528",
        "relative_path":"17B-197_2018_02_19_T15_59_16.097.tar",
        "checksum":"-1848259250",
        "checksum_type":"ngamsGenCrc32",
        "version":1,
        "size":108677120,
        "server":{"server":"nmngas01.aoc.nrao.edu:7777",
            "location":"DSOC",
            "cluster":"DSOC"
        }}

        Re-organize it to group files under servers so we can get all the
        files from each server in parallel.
        """

        servers = set([ngas_target.ngas_file.server for ngas_target in self.ngas_targets])

        result = {}
        for server in servers:
            targets = []
            for ngas_target in self.ngas_targets:
                target_copy = copy.deepcopy(ngas_target)
                if target_copy.ngas_file.server == server:
                    targets.append(target_copy)
            result[server] = targets

        return result

    def launch_retrievers(self) -> int:
        """
        Launch NGAS retrievers in parallel.

        :return: number of files retrieved
        """

        count = 0

        ngas_targets = []
        # VLA/VLBA files: all stored in NGAS
        for server_url, targets in self.files_by_server.items():

            for target in targets:
                ngas_targets.append(target)

            logger.debug(f"retrieving {len(targets)} files from {server_url}....")
            if self.dry_run:
                retriever = NgasNonRetriever(targets, self.force_overwrite)
            else:
                retriever = NgasRetriever(targets, self.force_overwrite)

            count += retriever.retrieve()

        count += self.retrieve_xml()

        return count

    def retrieve_xml(self) -> int:
        """
        Retrieve these SDM tables from the ALMA DB.

        :return:
        """
        count = 0
        for sdm in self.alma_sdms:
            sdm_file = OracleXmlRetriever().retrieve(sdm)
            self.validate_sdm(sdm_file, sdm)
            count += 1
        return count

    def validate_sdm(self, file: Path, alma_sdm: OracleXml):
        """
        Was this SDM table written? is it the right size?

        :param file: the SDM that was written to disk
        :param alma_sdm: the XML table retrieved from the ALMA DB

        :return:
        """
        if not file.is_file():
            raise FileNotFoundError(str(file))
        if file.stat().st_size != alma_sdm.size:
            raise FileErrorException(
                f"{file.name}: expected size "
                f"{alma_sdm.size} bytes, but size is "
                f"{file.stat().st_size} bytes"
            )
