""" Fixtures for datafetcher tests """
import json
import os
import tempfile
from pathlib import Path
from typing import List

# pylint: disable=E0401, R1721, W0621

import pytest
from datafetcher.locations_report import (
    LocationsReport,
    LocationsReportSchema,
    NgasTarget,
)

# conveniences used and re-used in tests
from datafetcher.utilities import ASDM_BIN_SUBDIR, get_file_info_from_json_file

FAKE_LOCATOR = "uid://evla/calibration/ibetyouthinkyourerealclever"


@pytest.fixture
def capo_profile_docker():
    os.environ["CAPO_PROFILE"] = "docker"


@pytest.fixture
def capo_profile_local():
    os.environ["CAPO_PROFILE"] = "local"


@pytest.fixture
def bad_server_json_file(resource_path_root) -> Path:
    """
    Get the test data location file for calibration.

    :return:
    """
    return resource_path_root / "location_files" / "VLA_BAD_SERVER.json"


@pytest.fixture
def cal_json_file(resource_path_root) -> Path:
    """
    Get the test data location file for calibration.

    :return:
    """
    return resource_path_root / "location_files" / "CALIBRATION.json"


@pytest.fixture
def cal_locations_report(cal_json_file) -> LocationsReport:
    """
    Grab all the file things from our calibration test .json

    :return:
    """
    locations_dict = get_file_info_from_json_file(cal_json_file)
    return LocationsReportSchema().load(locations_dict)


@pytest.fixture
def eb_17a_109_18468_locations_file(resource_path_root) -> Path:
    """
    Grab all the file things from one of our VLA test .jsons

    :return:
    """
    return resource_path_root / "location_files" / "17A-109_fg_18468.json"


@pytest.fixture
def eb_17a_109_18468_locations_report(
    eb_17a_109_18468_locations_file,
) -> LocationsReport:
    """
    Grab all the file things from our 17A-109....json

    :return: locations report for this file
    """
    locations_dict = get_file_info_from_json_file(eb_17a_109_18468_locations_file)
    return LocationsReportSchema().load(locations_dict)


@pytest.fixture
def vla_small_eb_locations_report(vla_small_locations_file):
    """
    Grab all the file things from our VLA_SMALL_EB.json

    :return: locations report for this file
    """
    locations_dict = get_file_info_from_json_file(vla_small_locations_file)
    assert len(locations_dict["files"]) == 44
    locations_report = LocationsReportSchema().load(locations_dict)
    assert len(locations_report.targets) == 44
    return locations_report


@pytest.fixture
def vla_small_locations_file(resource_path_root) -> Path:
    """
    Get our VLA small EB locations file.

    :return: locations file
    """
    return resource_path_root / "location_files" / "VLA_SMALL_EB.json"


@pytest.fixture
def alma_execblock_report(resource_path_root) -> Path:
    """
    Get our ALMA locations file.

    :return: locations file
    """
    return resource_path_root / "location_files" / "alma-execblock.json"


@pytest.fixture
def make_tempdir() -> Path:
    """
    Creates a new temporary working directory for each test.

    :return:
    """
    umask = os.umask(0o000)
    top_level = Path(tempfile.mkdtemp(prefix="datafetcher_test_", dir="/var/tmp"))
    os.umask(umask)
    return top_level


def write_fake_sdms(output_dir: Path, targets: List[NgasTarget], force: bool) -> List[Path]:
    """
    Write the files in in this locations report.

    :param output_dir: where to write the files
    :param targets: all the dope on the files to be written
    :param force: whether or not to force overwrite
    :return:
    """

    for target in targets:
        filename = target.relative_path
        if ASDM_BIN_SUBDIR in filename:
            continue

        file = output_dir / filename
        if file.exists() and not force:
            raise FileExistsError(f"{file}")
        with open(file, "w") as out:
            out.write(f"{target.size}")
    return [file for file in output_dir.iterdir() if file.is_file()]


def write_fake_exec_blocks(output_dir: Path, targets: List[NgasTarget], force: bool) -> List[Path]:
    """
    Write the files in in this locations report.

    :param output_dir: where to write the files
    :param targets: all the dope on the files to be written
    :param force: whether or not to force overwrite
    :return:
    """

    asdm_bin_dir = output_dir / ASDM_BIN_SUBDIR
    asdm_bin_dir.mkdir()

    for target in targets:
        if ASDM_BIN_SUBDIR in target.relative_path:
            relative_path = Path(target.relative_path)
            filename = relative_path.name
            file = asdm_bin_dir / filename
        else:
            file = output_dir / target.relative_path

        if file.exists() and not force:
            raise FileExistsError(f"{file}")
        with open(file, "w") as out:
            out.write(f"{target.size}")
    return [file for file in output_dir.iterdir() if file.is_file()]


def mock_retrieve_sdms(make_tempdir, targets: List[NgasTarget], force: bool) -> int:
    """
    Pretend to retrieve these SDMs. For each target, we'll write a
    little file containing the size of the real-file.

    :param make_tempdir: temporary test output dir
    :param targets: files to retrieve
    :param force: force-overwrite existing products, or not
    :return:
    """
    existing = [file for file in Path(make_tempdir).iterdir()]
    if force:
        assert len(existing) > 0
    retrieved = write_fake_sdms(make_tempdir, targets, force)
    return len(retrieved)


def mock_retrieve_exec_blocks(make_tempdir, targets: List[NgasTarget], force: bool) -> int:
    """
    Pretend to retrieve all the products in this execution block. For each
    target, we'll write a
    little file containing the size of the real-file.

    :param make_tempdir: temporary test output dir
    :param targets: files to retrieve
    :param force: force-overwrite existing products, or not
    :return:
    """
    existing = [file for file in Path(make_tempdir).iterdir()]
    if force:
        assert len(existing) > 0
    retrieved = write_fake_exec_blocks(make_tempdir, targets, force)
    return len(retrieved)


def write_fake_alma_locations_file(make_tempdir) -> Path:
    """
    Write a little ALMA locations report file.

    :return:
    """

    # create an SDM and a BDF
    bdf = {
        "ngas_file_id": "A002/Xe79954/X96e0",
        "subdirectory": "uid___A002_Xe79954_X8f3f",
        "relative_path": "ASDMBinary/uid___A002_Xe79954_X96e0",
        "checksum": "-250119162",
        "checksum_type": "ngamsGenCrc32",
        "version": 1,
        "size": 8275816,
        "server": {
            "server": "nangas14.cv.nrao.edu:7771",
            "location": "NAASC",
            "cluster": "NAASC",
        },
    }
    sdm = {
        "archive_uid": "uid://A002/Xe7b10b/X1",
        "table": "XML_CALPOINTINGTABLE_ENTITIES",
        "relative_path": "CalPointingTable.xml",
        "subdirectory": "uid___A002_Xe79954_X8f3f",
        "size": 6702,
    }

    outfile = Path(make_tempdir) / "locations.json"
    files = []
    files.append(sdm)
    files.append(bdf)
    aggregate_size = sdm["size"] + bdf["size"]
    json_dict = {"files": files, "aggregate_size": aggregate_size}
    json_str = json.dumps(json_dict)

    with open(outfile, "w") as out:
        out.write(json_str)

    return outfile


def we_are_in_docker() -> bool:
    return Path("/packages").exists()
