""" Test our fetch plans """

import pathlib
from unittest.mock import MagicMock

# pylint: disable=C0116, E0401, E0402, R0201

import pytest

from datafetcher.redesign.exceptions import FetchError
from datafetcher.redesign.fetch_plan import ParallelFetchPlan
from datafetcher.redesign.interfaces import FileFetcher, LocatedFile


class AlwaysSuccessfulFakeFetcher(FileFetcher):
    """ I can always fetch -- just ask me! """

    @property
    def file(self) -> LocatedFile:
        raise NotImplementedError

    def shares_resources_with(self, other: "FileFetcher") -> bool:
        raise NotImplementedError

    def do_fetch(self) -> pathlib.Path:
        return pathlib.Path.home()


class AlwaysFailingFakeFetcher(FileFetcher):
    """No matter how I try, I can't fetch. I'm pathetic. My life is
    meaningless."""

    @property
    def file(self) -> LocatedFile:
        raise NotImplementedError

    def shares_resources_with(self, other: "FileFetcher") -> bool:
        raise NotImplementedError

    def fetch(self):
        self.do_fetch()

    def do_fetch(self) -> pathlib.Path:
        raise FetchError("I always fail at this for some reason")


def test_parallel_plans_execute_all_fetchers(capsys):
    """
    Make a bunch of fetchers and confirm the fetch plan invokes them all.

    :param capsys:
    :return:
    """

    def with_fetchers_and_concurrency(fetcher_count: int, concurrency_level: int):
        # Prove that we always executed all the fetchers with no concurrency whatsoever
        fetchers = [MagicMock(wraps=AlwaysSuccessfulFakeFetcher()) for _ in range(fetcher_count)]
        ParallelFetchPlan(fetchers, concurrency=concurrency_level).fetch()
        for fetcher in fetchers:
            fetcher.fetch.assert_called()

    with_fetchers_and_concurrency(10, 1)
    with_fetchers_and_concurrency(16, 4)
    with_fetchers_and_concurrency(4, 16)
    capsys.readouterr()


def test_failures_always_found(capsys):
    """
    We should always fail when we expect to fail.

    :param capsys:
    :return:
    """
    fetchers = [MagicMock(wraps=AlwaysSuccessfulFakeFetcher()) for _ in range(7)]

    # no matter what concurrency level or where in the list we have an exception,
    # it should always cause the overall fetch plan to throw an exception
    for i in range(len(fetchers)):
        fetchers_copy = fetchers[:]
        fetchers_copy.insert(i, AlwaysFailingFakeFetcher())
        for concurrency_level in range(1, len(fetchers_copy)):
            with pytest.raises(FetchError):
                ParallelFetchPlan(fetchers_copy, concurrency=concurrency_level).fetch()

    capsys.readouterr()
