""" Tests for our various file fetchers """

# pylint: disable=C0103, E0401, E0402, W0621

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from datafetcher.redesign.exceptions import (
    RetryableFetchError,
    NormalModeFileExistsError,
    UnsatisfiableConstraint,
)
from datafetcher.redesign.fetchers import (
    RetryableFileFetcher,
    NgasDirectCopyFileFetcher,
    NgasStreamingFileFetcher,
    OracleXmlFetcher,
    ForceMode,
)
from datafetcher.redesign.interfaces import FileFetcher, LocatedFile
from datafetcher.redesign.locations import NgasFile, NgasServer, Location, Cluster, OracleXml


class FailNTimesFileFetcher(FileFetcher):
    """ A failin' fetcher. This dog don't fetch. """

    def __init__(self, n=3):
        super().__init__()
        self.failures = n

    @property
    def file(self) -> LocatedFile:
        """
        (Never called)

        :return:
        """

    def do_fetch(self) -> Path:
        """
        Try to fetch N times.

        :return:
        """
        if self.failures > 0:
            self.failures -= 1
            raise RetryableFetchError("testing failure " + str(self.failures))
        return Path.home()

    def shares_resources_with(self, other: "FileFetcher") -> bool:
        """
        (Never called)

        :return:
        """


@pytest.fixture
def ngas_file(tmpdir) -> NgasFile:
    """
    A fake NGAS file for our tests

    :param tmpdir:
    :return:
    """
    return NgasFile(
        "fake_ngas_id",
        tmpdir,
        "ASDMBinary",
        -1,
        "cksum",
        1,
        100,
        NgasServer("localhost", Location.DSOC, Cluster.DSOC),
    )


@pytest.fixture
def alma_file(tmpdir) -> OracleXml:
    """
    A fake ALMA SDM table for our tests

    :param tmpdir:
    :return:
    """
    return OracleXml("fake_uid", "fake_table", tmpdir, "relative_path", 100)


def test_force_mode(resource_path_root):
    """
    Does force mode overwrite an existing product?

    :param resource_path_root:
    :return:
    """
    # check files that do exist:
    assert (
        ForceMode.FORCE.check_destination(
            resource_path_root / "location_files" / "alma-execblock.json"
        )
        is None
    )
    with pytest.raises(NormalModeFileExistsError):
        ForceMode.NORMAL.check_destination(
            resource_path_root / "location_files" / "alma-execblock.json"
        )

    # check files that do not exist
    assert ForceMode.FORCE.check_destination(Path("/this/is/a/bogus/path")) is None
    assert ForceMode.NORMAL.check_destination(Path("/this/is/a/bogus/path")) is None


def test_ngas_direct_copy(ngas_file: NgasFile, tmpdir, capsys):
    """
    If we can direct copy, we should try to do so; if we can't, attempt to do so
     should throw exception

    :param ngas_file: what we're trying to fetch
    :param tmpdir: our temporary dir
    :param capsys:
    :return:
    """
    # make sure that if we could direct copy, we try to
    with patch("datafetcher.redesign.locations.NgasServer.can_direct_copy_to", return_value=True):
        with patch("datafetcher.redesign.ngas.NgasConnection.direct_copy", return_value=None):
            assert NgasDirectCopyFileFetcher(ngas_file).do_fetch() == Path(tmpdir) / "ASDMBinary"

    # contrarily, if we cannot, we should get an exception
    with patch("datafetcher.redesign.locations.NgasServer.can_direct_copy_to", return_value=False):
        with patch("datafetcher.redesign.ngas.NgasConnection.direct_copy", return_value=None):
            with pytest.raises(UnsatisfiableConstraint):
                NgasDirectCopyFileFetcher(ngas_file).do_fetch()

    capsys.readouterr()


def test_ngas_streaming(ngas_file: NgasFile, tmpdir, capsys):
    # make sure that if we could direct copy, we try to
    with patch("datafetcher.redesign.ngas.NgasConnection.stream", return_value=None):
        assert NgasStreamingFileFetcher(ngas_file).do_fetch() == Path(tmpdir / "ASDMBinary")

    capsys.readouterr()


def test_alma_xml(alma_file: OracleXml, tmpdir, capsys):
    mock_engine = MagicMock()
    mock_engine.execute.return_value = [["this is the file content"]]
    with patch(
        "datafetcher.redesign.fetchers.OracleXmlFetcher.get_engine", return_value=mock_engine
    ):
        assert OracleXmlFetcher(alma_file).do_fetch() == Path(tmpdir / "relative_path")

    capsys.readouterr()


# ----------------------------------------------------------------------------
#
#      R e t r y i n g   T e s t s
#
# ----------------------------------------------------------------------------


def test_retrying_succeeds(capsys):
    # two failures is OK
    with patch("time.sleep"):
        fail_twice = MagicMock(wraps=FailNTimesFileFetcher(2))
        result = RetryableFileFetcher(fail_twice).do_fetch()
        assert result == Path.home()
        assert fail_twice.do_fetch.call_count == 3
    capsys.readouterr()


def test_retrying_fails(capsys):
    # three failures is too many
    fail_twice = MagicMock(wraps=FailNTimesFileFetcher(3))
    with patch("time.sleep"):
        with pytest.raises(RetryableFetchError):
            RetryableFileFetcher(fail_twice).do_fetch()
    capsys.readouterr()
