""" Tests for the product fetcher """

# pylint: disable=C0103, E0401, E0402, E1136, R0201, R0903

from pathlib import Path
from typing import List, Union

import pytest

from datafetcher.redesign.fetchers import DryRunFakeFileFetcher, ForceMode
from datafetcher.redesign.interfaces import FetcherFactory, FileFetcher, LocatedFile, FetchPlan
from datafetcher.redesign.locations import FileLocator, ServiceLocator
from datafetcher.redesign.product_fetcher import FetchContext, CLIParam


class FakePlan(FetchPlan):
    """ A non-fetching fetch plan """

    def __init__(self, fetchers, concurrency):
        self.fetchers = fetchers
        self.concurrency = concurrency

    def fetch(self):
        # Let's not actually fetch anything today
        pass


class FakeFactory(FetcherFactory):
    """ Creates non-fetching fetchers """

    def fetch_ngas(self, file: LocatedFile) -> FileFetcher:
        return DryRunFakeFileFetcher(file)

    def fetch_oracle_xml(self, file: LocatedFile) -> FileFetcher:
        return DryRunFakeFileFetcher(file)

    def fetch_plan(
        self, fetchers: List[Union[FileFetcher, FetchPlan]], concurrency: int = 1
    ) -> FetchPlan:
        return FakePlan(fetchers, concurrency)


def test_plan_generation(resource_path_root):
    """
    Can we generate a test plan?

    :param resource_path_root:
    :return:
    """
    report = FileLocator(resource_path_root / "location_files" / "13B-014.json").locate()
    fake_factory = FakeFactory()
    fetchers = report.fetchers(fake_factory)

    # we should absolutely have the same number of fetchers as we have files
    assert len(fetchers) == len(report.files)

    # let's make a plan
    plan = FetchContext.calculate_plan(fake_factory, fetchers)

    # we should have the same number of fetchers in the plan as in the original report
    assert sum(len(subplan.fetchers) for subplan in plan.fetchers) == len(report.files)


def test_argument_parsing(capsys):
    """
    Can we parse the command-line arguments passed in?

    :param capsys:
    :return:
    """
    # ensure that various combinations do not work
    with pytest.raises(SystemExit):
        # must have an SPL or a file
        FetchContext.parse_commandline([])

    with pytest.raises(SystemExit):
        # cannot have an SPL and a file
        FetchContext.parse_commandline(
            [CLIParam.FILE.value, "foo", CLIParam.SPL.value, "uid://this/is/a/fakesy"]
        )

    # check the dry run value
    fc = FetchContext.parse_commandline([CLIParam.DRY.value, CLIParam.FILE.value, "foo"])
    assert fc.dry_run
    fc = FetchContext.parse_commandline([CLIParam.FILE.value, "foo"])
    assert not fc.dry_run

    # check the force value
    fc = FetchContext.parse_commandline([CLIParam.FORCE.value, CLIParam.FILE.value, "foo"])
    assert fc.force == ForceMode.FORCE
    fc = FetchContext.parse_commandline([CLIParam.FILE.value, "foo"])
    assert fc.force == ForceMode.NORMAL

    fc = FetchContext.parse_commandline([CLIParam.FILE.value, "foo"])
    assert isinstance(fc.locator, FileLocator)
    assert fc.locator.file == Path("foo")

    fc = FetchContext.parse_commandline([CLIParam.SPL.value, "uid://this/is/a/fakesy"])
    assert isinstance(fc.locator, ServiceLocator)
    assert fc.locator.spl == "uid://this/is/a/fakesy"

    fc = FetchContext.parse_commandline(
        [CLIParam.FILE.value, "foo", CLIParam.CONCURRENCY.value, "732"]
    )
    assert fc.concurrency == 732

    capsys.readouterr()
