""" Make sure the command-line parameters function as expected. """
from unittest.mock import MagicMock, patch

# pylint: disable=E0401, E0402, W0611, W0621

import pytest
from datafetcher.datafetcher import DataFetcher
from datafetcher.utilities import CLIParam

from .conftest import FAKE_LOCATOR

USAGE = "usage: "
_RUN_ALL = True

LOC_REPORT_BUILD_PATCH = "datafetcher.locations_report.LocationsReportBuilder.build"


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_no_args_displays_usage_and_exits(capsys):
    """
    If command-line parameters are omitted, print the "Usage" message, then exit.

    :param capsys:
    :return:
    """
    with pytest.raises(SystemExit):
        DataFetcher([])
    captured = capsys.readouterr()
    assert captured.err.startswith(USAGE)


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_help_flag_displays_usage_and_exits(make_tempdir, capsys):
    """
    The "-h" command-line parameter should print the "Usage" message, then exit.

    :return:
    """
    with pytest.raises(SystemExit):
        DataFetcher(["-h"])
    captured = capsys.readouterr()
    assert captured.out.startswith(USAGE)

    # longer help flag should be recognized and output_dir disregarded
    with pytest.raises(SystemExit):
        DataFetcher(["--help", "--output-dir", str(make_tempdir)])
    captured = capsys.readouterr()
    assert captured.out.startswith(USAGE)


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_locator_dry_run_recognized(cal_json_file):
    """
    Confirm that datafetcher recognizes "--dry-run" optional parameter with a product locator.

    :return:
    """

    args = [CLIParam.FILE.value, str(cal_json_file), CLIParam.DRY.value]

    fetcher = DataFetcher(args)
    assert fetcher.dry_run

    fetcher.run = MagicMock(name="run")
    fetcher.run()
    fetcher.run.assert_called_once()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_file_dry_run_recognized(cal_json_file):
    """
    Confirm that datafetcher recognizes "--dry-run" optional parameter with a locations file.

    :return:
    """
    args = [CLIParam.FILE.value, str(cal_json_file), CLIParam.DRY.value]
    fetcher = DataFetcher(args)
    assert fetcher.dry_run
    fetcher.run = MagicMock(name="run")
    fetcher.run()
    fetcher.run.assert_called_once()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_locator_plus_file_throws_missing_setting(capsys):
    """
    On the command line, product locator and location file are mutually
    exclusive options.

    :return:
    """
    args = [CLIParam.SPL.value, "locator", CLIParam.FILE.value, "file"]
    with pytest.raises(SystemExit):
        DataFetcher(args)
    captured = capsys.readouterr()
    assert captured.err.startswith(USAGE)


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_locator_no_force_default(cal_json_file):
    """
    If --force is not specified, it should default to False

    :return:
    """
    args = [CLIParam.FILE.value, str(cal_json_file)]
    datafetcher = DataFetcher(args)
    assert not datafetcher.force_overwrite


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_file_no_force_default(cal_json_file):
    """
    If --force is not specified, it should default to False

    :return:
    """
    args = [CLIParam.FILE.value, str(cal_json_file)]
    datafetcher = DataFetcher(args)
    assert not datafetcher.force_overwrite


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_locator_force_recognized():
    """
    Confirm that the force_overwrite option is recognized on the command line.

    :return:
    """
    args = [CLIParam.SPL.value, FAKE_LOCATOR, CLIParam.FORCE.value]
    with patch(LOC_REPORT_BUILD_PATCH):
        datafetcher = DataFetcher(args)
    assert datafetcher.force_overwrite


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_file_force_recognized(cal_json_file):
    """
    Confirm that the force_overwrite option is recognized on the command line.

    :return:
    """
    args = [CLIParam.FILE.value, str(cal_json_file), CLIParam.FORCE.value]
    datafetcher = DataFetcher(args)
    assert datafetcher.force_overwrite


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_locator_dry_plus_force_no_can_do():
    """
    The --dry-run and --force options make no sense together.

    :return:
    """
    args = [
        CLIParam.SPL.value,
        FAKE_LOCATOR,
        CLIParam.DRY.value,
        CLIParam.FORCE.value,
    ]
    with pytest.raises(SystemExit):
        DataFetcher(args)


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_file_dry_plus_force_no_can_do():
    """
    The --dry-run and --force options make no sense together.

    :return:
    """
    args = [
        CLIParam.FILE.value,
        "locations.json",
        CLIParam.DRY.value,
        CLIParam.FORCE.value,
    ]
    with patch("requests.sessions.Session.get"):
        with pytest.raises(SystemExit):
            DataFetcher(args)


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_direct_copy_flag_recognized_locator():
    """
    Confirm that datafetcher understands the --direct-copy flag.

    :return:
    """
    args = [CLIParam.SPL.value, FAKE_LOCATOR, CLIParam.DIRECT_COPY.value]
    with patch(LOC_REPORT_BUILD_PATCH):
        datafetcher = DataFetcher(args)
    assert datafetcher.direct_copy


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_streaming_flag_recognized_locator():
    """
    Confirm that datafetcher understands the --streaming flag.

    :return:
    """
    args = [CLIParam.SPL.value, FAKE_LOCATOR, CLIParam.STREAMING.value]
    with patch(LOC_REPORT_BUILD_PATCH):
        datafetcher = DataFetcher(args)
    assert datafetcher.streaming


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_direct_copy_flag_recognized_file(cal_json_file):
    """
    Confirm that datafetcher understands the --direct-copy flag.

    :return:
    """
    args = [CLIParam.FILE.value, str(cal_json_file), CLIParam.DIRECT_COPY.value]
    with patch(LOC_REPORT_BUILD_PATCH):
        datafetcher = DataFetcher(args)
    assert datafetcher.direct_copy


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_streaming_flag_recognized_file(cal_json_file):
    """
    Confirm that datafetcher understands the --streaming flag.

    :return:
    """
    args = [CLIParam.FILE.value, str(cal_json_file), CLIParam.STREAMING.value]
    datafetcher = DataFetcher(args)
    assert datafetcher.streaming


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_stream_plus_copy_throws_missing_setting(cal_json_file):
    """
    The --direct-copy and --streaming options are mutually exclusive.

    :param cal_json_file: locations report file
    :return:
    """
    locator_args = [
        CLIParam.SPL.value,
        FAKE_LOCATOR,
        CLIParam.STREAMING.value,
        CLIParam.DIRECT_COPY.value,
    ]
    file_args = [
        CLIParam.FILE.value,
        str(cal_json_file),
        CLIParam.STREAMING.value,
        CLIParam.DIRECT_COPY.value,
    ]
    for args in locator_args, file_args:
        with pytest.raises(SystemExit):
            DataFetcher(args)
