""" Tests for -function- (not the CLI) of datafetcher command-line options """
import os
from pathlib import Path
from unittest.mock import patch, MagicMock

# pylint: disable=E0401, E0402, W0611, W0613, W0621

import pytest
from datafetcher.errors import NGASServiceErrorException
from datafetcher.locations_report import LocationsReportBuilder

from datafetcher.utilities import CLIParam, LUSTRE, get_file_info_from_json_file
from datafetcher.datafetcher import DataFetcher

from .conftest import (
    FAKE_LOCATOR,
    mock_retrieve_sdms,
    vla_small_eb_locations_report,
    write_fake_alma_locations_file,
)

_RUN_ALL = True


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_no_write_on_dry_run(cal_json_file):
    """
    Make sure we're not writing files if dry_run is set.

    :return:
    """
    args = [CLIParam.FILE.value, str(cal_json_file), CLIParam.DRY.value]
    datafetcher = DataFetcher(args)
    assert datafetcher.dry_run

    with patch("datafetcher.ngas_retriever.NgasRetriever.__init__") as mock:
        datafetcher.run = MagicMock(name="run")
        datafetcher.run()
        mock.assert_not_called()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_no_overwrite_without_force(
    make_tempdir, vla_small_locations_file, vla_small_eb_locations_report
):
    """
    If a product already exists in the output directory, balk.

    :return:
    """

    # create a file in the output dir
    existing_file = Path(make_tempdir) / "ASDM.xml"
    with open(existing_file, "w") as out:
        out.write("He's -not- the Messiah -- he's a very naughty boy!\n")
    assert existing_file.exists()

    args = [CLIParam.FILE.value, str(vla_small_locations_file)]
    datafetcher = DataFetcher(args)
    assert not datafetcher.force_overwrite

    # since --force is not specified, trying to overwrite existing file
    # should tank
    with pytest.raises(FileExistsError):
        with patch(
            "datafetcher.ngas_retriever.NgasRetriever.retrieve",
            return_value=mock_retrieve_sdms(
                make_tempdir, vla_small_eb_locations_report.targets, force=False
            ),
        ):
            datafetcher.run()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_overwrites_existing_file_when_force(
    make_tempdir, vla_small_locations_file, vla_small_eb_locations_report
):
    """
    If a product already exists in the output directory, it should still be
    retrieved when --force flag is passed to datafetcher.

    :return:
    """
    # create a file in the output dir
    existing_file = Path(make_tempdir) / "ASDM.xml"
    with open(existing_file, "w") as out:
        out.write("He's -not- the Messiah -- he's a very naughty boy!\n")
    assert existing_file.exists()
    assert existing_file.stat().st_size > 4

    args = [CLIParam.FILE.value, str(vla_small_locations_file)]
    datafetcher = DataFetcher(args)
    assert not datafetcher.force_overwrite

    # since --force is specified, existing file should be overwritten
    with patch(
        "datafetcher.ngas_retriever.NgasRetriever.retrieve",
        return_value=mock_retrieve_sdms(
            make_tempdir, vla_small_eb_locations_report.targets, force=True
        ),
    ):
        datafetcher.run()
        assert existing_file.stat().st_size == 4


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_direct_copy_flag_forces_copy(
    monkeypatch, make_tempdir, cal_locations_report, cal_json_file
):
    """
    If --direct-copy is specified on the command line AND direct copy is
    feasible, use the direct-copy plugin when fetching from NGAS.

    :return:
    """

    def mock_locations_from_service(*args, **kwargs):
        return get_file_info_from_json_file(cal_json_file)

    os.chdir(Path(make_tempdir))
    locator_args = [CLIParam.SPL.value, FAKE_LOCATOR, CLIParam.DIRECT_COPY.value]
    file_args = [CLIParam.FILE.value, str(cal_json_file), CLIParam.DIRECT_COPY.value]
    monkeypatch.setattr(
        LocationsReportBuilder,
        "get_locations_from_service",
        mock_locations_from_service,
    )
    with patch("time.sleep"):
        with patch("datafetcher.ngas_retriever.DryCopier.send_request") as mock_copy:
            with patch("datafetcher.ngas_retriever.DryStreamer.send_request") as mock_stream:
                for args in locator_args, file_args:
                    datafetcher = DataFetcher(args)
                    # we're not on lustre, so we should get an error
                    with pytest.raises(NGASServiceErrorException):
                        datafetcher.run()
                        mock_copy.assert_called()
                        mock_stream.assert_not_called()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_stream_flag_forces_stream(make_tempdir, cal_json_file):
    """
    If --streaming is specified on the command line, always request streaming
    from NGAS, even if direct-copy would be faster and more efficient.

    :return:
    """

    output_dir = Path(make_tempdir) / LUSTRE
    output_dir.mkdir()
    os.chdir(output_dir)

    locator_args = [CLIParam.SPL.value, FAKE_LOCATOR, CLIParam.STREAMING.value]
    file_args = [CLIParam.FILE.value, str(cal_json_file), CLIParam.STREAMING.value]
    with patch(
        "datafetcher.locations_report.LocationsReportBuilder.get_locations_from_service",
        return_value=mock_locations_from_service(cal_json_file),
    ):
        with patch("time.sleep"):
            with patch("datafetcher.ngas_retriever.NgasCopier.send_request") as mock_copy:
                with patch("datafetcher.ngas_retriever.NgasStreamer.send_request") as mock_stream:
                    for args in locator_args, file_args:
                        datafetcher = DataFetcher(args)
                        datafetcher.run()
                        mock_copy.assert_not_called()
                        mock_stream.assert_called()


@pytest.mark.skipif(not _RUN_ALL, reason="debugging sth else just now")
def test_direct_copy_error_if_no_can_do(cal_json_file):
    """
    If direct-copy is specified on the command line but is not possible
    because we're not running on /lustre, we should get an appropriate error.

    :return:
    """

    args = [CLIParam.SPL.value, FAKE_LOCATOR, CLIParam.DIRECT_COPY.value]
    with patch(
        "datafetcher.locations_report.LocationsReportBuilder.get_locations_from_service",
        return_value=mock_locations_from_service(cal_json_file),
    ):
        datafetcher = DataFetcher(args)
        with pytest.raises(NGASServiceErrorException):
            datafetcher.run()


def mock_locations_from_service(cal_json_file):
    """

    Pretend to get a calibration from archiveService

    :param cal_json_file: our calibration .json
    :return:
    """
    return get_file_info_from_json_file(cal_json_file)
