""" Proof of concept: uses marshmallow to load locations report JSON
    directly into a ClassyLocationsReport!
"""

# pylint: disable=E0401, E0402, W0611, W0621

from pathlib import Path
from typing import NamedTuple
from unittest.mock import patch, MagicMock

import requests_mock

from datafetcher.redesign.interfaces import LocationReport
from datafetcher.redesign.locations import (
    FileLocator,
    OracleXml,
    NgasFile,
    NgasServer,
    Location,
    Cluster,
    ServiceLocator,
)


def test_loads_simple_locations_report(cal_json_file: Path):
    """
    Can marshmallow load an entire JSON locations report?

    :return:
    """
    report = FileLocator(cal_json_file).locate()
    assert isinstance(report, LocationReport)


def test_loads_complex_locations_report(eb_17a_109_18468_locations_file):
    """
    Can marshmallow load an entire JSON locations report?

    :return:
    """
    locations_rpt = FileLocator(eb_17a_109_18468_locations_file).locate()
    assert isinstance(locations_rpt, LocationReport)

    files = locations_rpt.files
    assert len(files) == 79
    assert locations_rpt.aggregate_size == 45572444823


def test_loads_alma_locations_report(alma_execblock_report):
    """
    Can marshmallow load an entire JSON locations report?

    :return:
    """
    locations_rpt = FileLocator(alma_execblock_report).locate()
    assert isinstance(locations_rpt, LocationReport)

    assert len(locations_rpt.files) == 1099
    assert len([file for file in locations_rpt.files if isinstance(file, OracleXml)]) == 36
    assert len([file for file in locations_rpt.files if isinstance(file, NgasFile)]) == 1063
    assert locations_rpt.aggregate_size == 4119667881


class FakeDatafetcherSettings(NamedTuple):
    defaultThreadsPerHost: int
    executionSite: str
    locatorServiceUrlPrefix: str


def test_loads_from_service(eb_17a_109_18468_locations_file: Path):
    """
    Can we access a web service and retrieve a location report?

    :param eb_17a_109_18468_locations_file: report file to use as a fake
    :return:
    """
    settings = FakeDatafetcherSettings(4, "DSOC", "http://localhost/location?locator=")
    with patch("datafetcher.redesign.locations.CapoConfig") as capo:
        capo.return_value = MagicMock()
        capo.return_value.settings.return_value = settings
        with requests_mock.Mocker() as mock:
            mock.get(
                "http://localhost/location?locator=pretend-uid",
                text=eb_17a_109_18468_locations_file.read_text(),
            )
            locations_rpt = ServiceLocator("pretend-uid").locate()
            assert len(locations_rpt.files) == 79
            assert locations_rpt.aggregate_size == 45572444823


def test_direct_copy_detection():
    """
    Ensure that we properly detect various scenarios when we can use
    the direct copy plugin.

    :return:
    """
    aoc = Path("/lustre/aoc/foo")
    naasc = Path("/lustre/naasc/foo")
    home = Path("/home/nobody/foo")

    # there are two cases we can direct copy to, currently: DSOC -> DSOC
    # and NAASC -> NAASC (with cluster=DSOC)
    assert NgasServer("", Location.DSOC, Cluster.DSOC).can_direct_copy_to(aoc)
    assert NgasServer("", Location.NAASC, Cluster.DSOC).can_direct_copy_to(naasc)

    # all the other permutations we cannot: NAASC -> AOC, AOC -> NAASC, NAASC -> NAASC
    assert not NgasServer("", Location.NAASC, Cluster.DSOC).can_direct_copy_to(aoc)
    assert not NgasServer("", Location.DSOC, Cluster.NAASC).can_direct_copy_to(aoc)
    assert not NgasServer("", Location.DSOC, Cluster.NAASC).can_direct_copy_to(naasc)
    assert not NgasServer("", Location.NAASC, Cluster.NAASC).can_direct_copy_to(naasc)
    assert not NgasServer("", Location.NAASC, Cluster.NAASC).can_direct_copy_to(aoc)

    # and of course, we can never direct copy to your house
    for location in Location:
        for cluster in Cluster:
            assert not NgasServer("", location, cluster).can_direct_copy_to(home)
