""" Proof of concept: retrieve ALMA BDF and SDMs """

import json
import logging
import os
from pathlib import Path

# pylint: disable=E0401, E0402, W0106, W0611, W0613, W0621
from typing import List
from unittest.mock import patch

from datafetcher.ngas_retriever import NgasRetriever
from datafetcher.oracle_xml_retriever import OracleXmlRetriever
from datafetcher.datafetcher import DataFetcher
from datafetcher.errors import FileErrorException

import pytest

from datafetcher.locations_report import (
    NgasFileSchema,
    NgasFile,
    NgasTarget,
    OracleXml,
)
from datafetcher.retrieve_launcher import ParallelRetrieverLauncher
from datafetcher.utilities import (
    LOG_FORMAT,
    CLIParam,
    get_file_info_from_json_file,
    ASDM_BIN_SUBDIR,
)

from .conftest import we_are_in_docker, write_fake_alma_locations_file

logger = logging.getLogger(__name__)
formatter = logging.Formatter(LOG_FORMAT)
handler = logging.StreamHandler()
handler.setFormatter(fmt=formatter)
logger.addHandler(handler)

_RUN_ALL = True


@pytest.mark.skipif(not _RUN_ALL, reason="worked last time I tried it")
@pytest.mark.skipif(we_are_in_docker(), reason="must be run locally")
def test_understands_sdm_vs_bdf(make_tempdir, capo_profile_local):
    """
    Given an ALMA target, locations should choose correct what to do for
    an SDM (get the XML from ALMA DB) vs a BDF (retrieve it from NGAS)

    :return:
    """
    locations_file = write_fake_alma_locations_file(make_tempdir)

    args = [CLIParam.FILE.value, str(locations_file)]
    datafetcher = DataFetcher(args)
    with patch(
        "datafetcher.retrieve_launcher.ParallelRetrieverLauncher.retrieve_xml",
        return_value=6702,
    ) as mock_sdm_retrieve:
        with patch(
            "datafetcher.ngas_retriever.NgasRetriever.retrieve", return_value=8275816
        ) as mock_bdf_retrieve:
            datafetcher.run()
            mock_sdm_retrieve.assert_called_once()
            mock_bdf_retrieve.assert_called_once()


@pytest.mark.skipif(not _RUN_ALL, reason="worked last time I tried it")
@pytest.mark.skipif(we_are_in_docker(), reason="must be run locally")
# heads up: capo_profile_local -is- used
def test_fetches_sdm_from_alma_locations_report(
    monkeypatch, capo_profile_local, make_tempdir, alma_execblock_report
):
    """
    Fake-retrieve the SDMs from an ALMA locations report file

    :return:
    """

    expected_size = 1280

    def mock_launch_retrievers(*args):
        """
        Pretend to retrieve Oracle XML.

        :return:
        """
        alma_sdms = alma_xmls(alma_execblock_report)
        sdm = [
            sdm
            for sdm in alma_sdms
            if sdm.archive_uid == "uid://A002/Xe7b108/X1"
            and sdm.table == "XML_CALATMOSPHERETABL_ENTITIES"
        ][0]
        assert sdm.size == expected_size
        file = Path(make_tempdir) / sdm.relative_path
        to_write = ""
        for _ in range(sdm.size):
            to_write += "0"
        with open(file, "w") as out:
            out.write(to_write)

        assert file.is_file()
        file_size = file.stat().st_size
        assert file_size == expected_size

    monkeypatch.setattr(ParallelRetrieverLauncher, "launch_retrievers", mock_launch_retrievers)
    args = [CLIParam.FILE.value, str(Path(alma_execblock_report))]
    os.chdir(Path(make_tempdir))
    datafetcher = DataFetcher(args)
    datafetcher.run()

    fake_sdm_file = Path(make_tempdir) / "CalAtmosphereTable.xml"
    assert fake_sdm_file.is_file()
    assert fake_sdm_file.stat().st_size == expected_size


# heads up: capo_profile_local -is- used
@pytest.mark.skipif(not _RUN_ALL, reason="worked last time I tried it")
@pytest.mark.skipif(we_are_in_docker(), reason="must be run locally")
def test_fetches_bdf_from_alma_locations_report(
    monkeypatch, capo_profile_local, make_tempdir, alma_execblock_report
):
    """
    Fake-retrieve a small BDF from an ALMA locations report file

    :return:
    """

    def mock_launch_retrievers(*args):
        """
        Pretend to fetch the BDFs in an ALMA locations report
        :return:
        """
        bdfs = alma_bdfs(alma_execblock_report)
        bdf = [bdf for bdf in bdfs if bdf.ngas_file_id == "A002/Xe79954/X8f5b"][0]

        bin_dir = Path(make_tempdir) / ASDM_BIN_SUBDIR
        bin_dir.mkdir()

        filename = bdf.relative_path.replace(ASDM_BIN_SUBDIR + "/", "")
        file = bin_dir / filename
        with open(file, "w") as out:
            [out.write("0") for _ in range(bdf.size)]

        assert file.stat().st_size == 3199

    monkeypatch.setattr(ParallelRetrieverLauncher, "launch_retrievers", mock_launch_retrievers)
    args = [CLIParam.FILE.value, str(Path(alma_execblock_report))]
    os.chdir(Path(make_tempdir))

    # assertion is in the mock function
    DataFetcher(args).run()


@pytest.mark.skipif(we_are_in_docker(), reason="must be run locally")
@pytest.mark.skipif(not _RUN_ALL, reason="worked last time I tried it")
def test_fetches_complete_alma_locations_report(
    monkeypatch, make_tempdir, capo_profile_local, alma_execblock_report
):
    """
    Fake-retrieve every SDM and BDF in ALMA locations report.

    :return:
    """

    def mock_launch_retrievers(*args):
        """
        Pretend to fetch all the files in an ALMA locations report
        :return:
        """

        # Pretend to fetch the BDFs
        bdfs = alma_bdfs(alma_execblock_report)
        assert len(bdfs) == 1060
        bin_dir = Path(make_tempdir) / ASDM_BIN_SUBDIR
        bin_dir.mkdir()
        count = 0
        for bdf in bdfs:
            count += 1
            write_fake_bdf(bin_dir, bdf.ngas_file)

        # Pretend to fetch the SDMs
        sdms = alma_xmls(alma_execblock_report)
        assert len(sdms) == 36
        for sdm in sdms:
            count += 1
            write_fake_oracle_xml(make_tempdir, sdm)

        assert count == len(sdms) + len(bdfs)
        return count

    monkeypatch.setattr(ParallelRetrieverLauncher, "launch_retrievers", mock_launch_retrievers)
    args = [CLIParam.FILE.value, str(Path(alma_execblock_report))]
    os.chdir(Path(make_tempdir))
    datafetcher = DataFetcher(args)
    datafetcher.run()


@pytest.mark.skipif(we_are_in_docker(), reason="must be run locally")
@pytest.mark.skipif(not _RUN_ALL, reason="worked last time I tried it")
def test_missing_sdm_throws_exception(monkeypatch):
    """
    If we can't find an SDM table for this archive_uid, we should howl
    like a two-year-old pitching a tantrum in the supermarket

    :return:
    """

    def mock_retrieve(*args):
        raise FileErrorException()

    file_spec = {
        "archive_uid": "uid://A002/Xe7b10d/X1",
        "table": "XML_CALWVRTABLE_ENTITIES",
        "relative_path": "CalWVRTable.xml",
        "subdirectory": "uid___A002_Xe79954_X8f3f",
        "size": 24345,
    }
    json_str = json.dumps(file_spec)
    sdm = NgasFileSchema().loads(json_str)

    monkeypatch.setattr(OracleXmlRetriever, "retrieve", mock_retrieve)
    with pytest.raises(FileErrorException):
        OracleXmlRetriever().retrieve(sdm)


@pytest.mark.skipif(not _RUN_ALL, reason="worked last time I tried it")
def test_missing_bdf_throws_file_err(monkeypatch):
    """
    If we can't find this BDF, we should make like a politician on Twitter
    denying sexual misconduct allegations

    :param monkeypatch:
    :return:
    """

    def mock_retrieve(*args):
        raise FileErrorException()

    file_spec = {
        "ngas_file_id": "A002/Xe79954/X9007",
        "subdirectory": "uid___A002_Xe79954_X8f3f",
        "relative_path": "ASDMBinary/uid___A002_Xe79954_X9007",
        "checksum": "1262621966",
        "checksum_type": "ngamsGenCrc32",
        "version": 1,
        "size": 23639454,
        "server": {
            "server": "nangas12.cv.nrao.edu:7772",
            "location": "NAASC",
            "cluster": "NAASC",
        },
    }
    json_str = json.dumps(file_spec)
    ngas_file = NgasFileSchema().loads(json_str)
    target = NgasTarget(ngas_file, "NAASC")
    monkeypatch.setattr(NgasRetriever, "retrieve", mock_retrieve)
    with pytest.raises(FileErrorException):
        NgasRetriever([target], False).retrieve()


# ---------------------"
#  U T I L I T I E S   "
# ---------------------"


def write_fake_oracle_xml(make_tempdir, sdm: OracleXml):
    """
    Pretend to write an ALMA SDM table.

    :param make_tempdir: temporary directory
    :param sdm: ALMA SDM
    :return:
    """
    file = Path(make_tempdir) / sdm.relative_path
    with open(file, "w") as out:
        out.write(f"{len(sdm)}")
    return file


def write_fake_bdf(bin_dir: Path, ngas_file: NgasFile):
    """
    Pretend to write an ALMA BDF.

    :param bin_dir: ASDMBinary/ in our tempdir
    :param ngas_file: ALMA BDF
    :return:
    """
    rel_path = ngas_file.relative_path
    file = bin_dir / rel_path.replace(ASDM_BIN_SUBDIR + "/", "")
    with open(file, "w") as out:
        out.write(f"{ngas_file.size}")
    return file


def alma_bdfs(alma_execblock_report) -> List[NgasTarget]:
    """
    Fake-retrieve BDFs from ALMA locations report

    :param alma_execblock_report: our faux ALMA locations report
    :return:
    """

    locations_dict = get_file_info_from_json_file(Path(alma_execblock_report))
    bdf_targets = []
    schema = NgasFileSchema()
    for file in locations_dict["files"]:
        if ASDM_BIN_SUBDIR in file["relative_path"]:
            # if "ngas_file_id" in file.keys():
            json_str = json.dumps(file)
            ngas_file = schema.loads(json_str)
            bdf_targets.append(NgasTarget(ngas_file, "NAASC"))

    return bdf_targets


def alma_xmls(alma_execblock_report) -> List[OracleXml]:
    """
    Fake-retrieve SDMs as OracleXml from ALMA locations report

    :param alma_execblock_report: our faux ALMA locations report
    :return:
    """

    locations_dict = get_file_info_from_json_file(Path(alma_execblock_report))
    xmls = []
    schema = NgasFileSchema()
    for file in locations_dict["files"]:
        if "table" in file.keys():
            json_str = json.dumps(file)
            alma_sdm = schema.loads(json_str)
            xmls.append(alma_sdm)

    return xmls
